/********************************************************************
*
*   File: timecode.cpp
*   Purpose: SSL timecode signal analysis
*
*   From kblplayer src, released under GPL. kblsalat 2011
*
********************************************************************/

#include "include.h"

//*******************************************************************
// Constructor & Destructor

CTimecode::CTimecode(DWORD dwBufSize, DOUBLE fSampleRate)
{
    this->Reset();

    this->m_fSampleRate = fSampleRate;

    // Assume SSL CD Timecode

    this->m_fDefaultFreq = 1000.0;
    this->m_wDefaultPhase = PHASE_RIGHT_LEFT;

    this->m_bLearn = FALSE;
}

CTimecode::~CTimecode(void)
{

}

//*******************************************************************
// Analyzes the signal

VOID CTimecode::Analyze(PTIMECODE_INFO pInfo, PFLOAT pBuffer, DWORD dwLength)
{
    // Find peaks

    this->m_dwPeaks = {0, 0};

    this->m_dwLowPeaks = 0;
    this->m_dwHighPeaks = 0;
    this->m_dwZeroPeaks = 0;

    WORD wChannel = 0;

    DWORD dwFirstZeroPeak = 1025;
    DWORD dwLastHighPeak[2] = {1025, 1025};
    DWORD dwLastLowPeak[2] = {1025, 1025};
    DWORD dwLastZeroPeak[2] = {1025, 1025};

    BOOL bAmplitudeThresholdReached = FALSE;

    for (DWORD i=0; i<(dwLength/4)-4; i++)
    {
        // Zero crossing?

        if (pBuffer[i] >= 0.0f && pBuffer[i+2] <= 0.0f)
        {
            if (dwLastZeroPeak[wChannel] == 1025)
            {
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].dwPosition = i;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wAmplitudeChange = AMPLITUDE_DOWN;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wType = PEAK_ZERO;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].fTime = this->m_Time + ((i-wChannel) / this->m_fSampleRate);

                dwLastZeroPeak[wChannel] = this->m_dwPeaks[wChannel];

                if (wChannel == 0)
                {
                    dwFirstZeroPeak = this->m_dwPeaks[0];
                    this->m_dwZeroPeaks++;
                }

                this->m_dwPeaks[wChannel]++;
            }
            else if (i > this->m_Peaks[wChannel][dwLastZeroPeak[wChannel]].dwPosition + 16)
            {
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].dwPosition = i;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wAmplitudeChange = AMPLITUDE_DOWN;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wType = PEAK_ZERO;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].fTime = this->m_Time + ((i-wChannel) / this->m_fSampleRate);

                dwLastZeroPeak[wChannel] = this->m_dwPeaks[wChannel];

                if (wChannel == 0)
                {
                    this->m_dwZeroPeaks++;
                }

                this->m_dwPeaks[wChannel]++;
            }
        }
        else if (pBuffer[i] <= 0.0f && pBuffer[i+2] >= 0.0f)
        {
            if (dwLastZeroPeak[wChannel] == 1025)
            {
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].dwPosition = i;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wAmplitudeChange = AMPLITUDE_UP;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wType = PEAK_ZERO;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].fTime = this->m_Time + ((i-wChannel) / this->m_fSampleRate);

                dwLastZeroPeak[wChannel] = this->m_dwPeaks[wChannel];

                if (wChannel == 0)
                {
                    dwFirstZeroPeak = this->m_dwPeaks[wChannel];
                    this->m_dwZeroPeaks++;
                }

                this->m_dwPeaks[wChannel]++;
            }
            else if (i > this->m_Peaks[wChannel][dwLastZeroPeak[wChannel]].dwPosition + 16)
            {
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].dwPosition = i;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wAmplitudeChange = AMPLITUDE_UP;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wType = PEAK_ZERO;
                this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].fTime = this->m_Time + ((i-wChannel) / this->m_fSampleRate);

                dwLastZeroPeak[wChannel] = this->m_dwPeaks[wChannel];

                if (wChannel == 0)
                {
                    this->m_dwZeroPeaks++;
                }

                this->m_dwPeaks[wChannel]++;
            }
        }

        // High peak?

        else if (pBuffer[i] <= pBuffer[i+2] && pBuffer[i+2] >= pBuffer[i+4])
        {
            if (pBuffer[i] >= 0.01f)
            {
                bAmplitudeThresholdReached = TRUE;

                if (dwLastHighPeak[wChannel] == 1025)
                {
                    dwLastHighPeak[wChannel] = this->m_dwPeaks[wChannel];

                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].dwPosition = i;
                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wType = PEAK_HIGH;
                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].fTime = this->m_Time + ((i-wChannel) / this->m_fSampleRate);

                    this->m_dwPeaks[wChannel]++;
                }
                else if (i > this->m_Peaks[wChannel][dwLastHighPeak[wChannel]].dwPosition + 16)
                {
                    dwLastHighPeak[wChannel] = this->m_dwPeaks[wChannel];

                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].dwPosition = i;
                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wType = PEAK_HIGH;
                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].fTime = this->m_Time + ((i-wChannel) / this->m_fSampleRate);

                    this->m_dwPeaks[wChannel]++;
                }
            }
        }

        // Low peak?

        else if (pBuffer[i] >= pBuffer[i+2] && pBuffer[i+2] <= pBuffer[i+4])
        {
            if (pBuffer[i] <= -0.01f)
            {
                bAmplitudeThresholdReached = TRUE;

                if (dwLastLowPeak[wChannel] == 1025)
                {
                    dwLastLowPeak[wChannel] = this->m_dwPeaks[wChannel];

                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].dwPosition = i;
                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wType = PEAK_LOW;
                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].fTime = this->m_Time + ((i-wChannel) / this->m_fSampleRate);

                    this->m_dwPeaks[wChannel]++;
                }
                else if (i > this->m_Peaks[wChannel][dwLastLowPeak[wChannel]].dwPosition + 16)
                {
                    dwLastLowPeak[wChannel] = this->m_dwPeaks[wChannel];

                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].dwPosition = i;
                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].wType = PEAK_LOW;
                    this->m_Peaks[wChannel][this->m_dwPeaks[wChannel]].fTime = this->m_Time + ((i-wChannel) / this->m_fSampleRate);

                    this->m_dwPeaks[wChannel]++;
                }
            }
        }

        wChannel = !wChannel;
    }

    // Paused?

    if (this->m_dwZeroPeaks <= 4 || !bAmplitudeThresholdReached)
    {
        pInfo->bPaused = TRUE;
        return;
    }
    else pInfo->bPaused = FALSE;

    // find the last "valid" zero crossing. the amplitude direction has to match the one of the first crossing to get a full wave cycle

    dwLastZeroPeak[0] = dwFirstZeroPeak;
    DWORD dwCount = this->m_dwPeaks[0]-1;
    DWORD dwSub = 0;

    while (dwLastZeroPeak[0] == dwFirstZeroPeak)
    {
        if (this->m_Peaks[0][dwCount].wType == PEAK_ZERO)
        {
            if (this->m_Peaks[0][dwCount].wAmplitudeChange == this->m_Peaks[0][dwFirstZeroPeak].wAmplitudeChange)
            {
                dwLastZeroPeak[0] = dwCount;
                break;
            }

            dwSub++;
        }
        if (dwCount == 0)
        {
            return;
        }
        else
        {
            dwCount--;
        }
    }

    // get direction by analyzing the phase shift

    DWORD dwHighLeft[2] = {1025, 1025};
    DWORD dwHighRight;

    for (DWORD i=0; i<this->m_dwPeaks[0]; i++)
    {
        if (this->m_Peaks[0][i].wType == PEAK_HIGH)
        {
            if (dwHighLeft[0] == 1025) dwHighLeft[0] = i;
            else
            {
                dwHighLeft[1] = i;
                break;
            }
        }
    }

    for (DWORD i=0; i<this->m_dwPeaks[1]; i++)
    {
        if (this->m_Peaks[1][i].wType == PEAK_HIGH)
        {
            if (this->m_Peaks[0][dwHighLeft[0]].dwPosition < this->m_Peaks[1][i].dwPosition)
            {
                dwHighRight = i;
                break;
            }
        }
    }

    if (this->m_bLearn)
    {
       if (this->m_Peaks[1][dwHighRight].dwPosition - this->m_Peaks[0][dwHighLeft[0]].dwPosition > (this->m_Peaks[0][dwHighLeft[1]].dwPosition - this->m_Peaks[0][dwHighLeft[0]].dwPosition) / 2)
       {
           this->m_wDefaultPhase = PHASE_RIGHT_LEFT;
       }
       else this->m_wDefaultPhase = PHASE_LEFT_RIGHT;
    }

    if (this->m_Peaks[1][dwHighRight].dwPosition - this->m_Peaks[0][dwHighLeft[0]].dwPosition > (this->m_Peaks[0][dwHighLeft[1]].dwPosition - this->m_Peaks[0][dwHighLeft[0]].dwPosition) / 2)
    {
        if (this->m_wDefaultPhase == PHASE_RIGHT_LEFT) pInfo->bDirection = FORWARD;
        else pInfo->bDirection = BACKWARD;
    }
    else
    {
        if (this->m_wDefaultPhase == PHASE_RIGHT_LEFT) pInfo->bDirection = BACKWARD;
        else pInfo->bDirection = BACKWARD;
    }

    // calculate pitch

    DOUBLE fFreq = ((this->m_dwZeroPeaks - dwSub) / 2) / ((this->m_Peaks[0][dwLastZeroPeak[0]].fTime - this->m_Peaks[0][dwFirstZeroPeak].fTime) / 2.0);

    pInfo->fVelocity = (FLOAT)floorl((fFreq - this->m_fDefaultFreq) * 10.0 + 0.5) / 100.0;

    //printf("Pitch: %f%%\n", pInfo->fVelocity);

    // Update timer

    this->m_Time += dwLength / ONE_SAMPLE_IN_SECONDS;
}

//*******************************************************************
// Resets the timer

VOID CTimecode::Reset(void)
{
    ZeroMemory(&this->m_Peaks, 2*1024*sizeof(PEAK));
    this->m_dwPeaks = {0, 0};

    this->m_dwHighPeaks = 0;
    this->m_dwLowPeaks = 0;
    this->m_dwZeroPeaks = 0;

    this->m_Time = 0.0;
}

//*******************************************************************
// Sets the learn flag

VOID CTimecode::Learn(void)
{
    this->m_bLearn = TRUE;
}
