/********************************************************************
*
*   File: timecode.h
*   Purpose: SSL timecode signal analysis
*
*   From kblplayer src, released under GPL. kblsalat 2011
*
********************************************************************/

#ifndef _TIMECODE_H_
#define _TIMECODE_H_

#define PEAK_ZERO 2
#define PEAK_HIGH 1
#define PEAK_LOW 0

#define AMPLITUDE_DOWN 0
#define AMPLITUDE_UP 1

#define PHASE_RIGHT_LEFT 0 // right peaks before left
#define PHASE_LEFT_RIGHT 1 // left peaks before right

#define ONE_SAMPLE_IN_SECONDS (8.0 / this->m_fSampleRate)

typedef struct _PEAK
{
    WORD wType;
    DWORD dwPosition;
    WORD wAmplitudeChange;
    DOUBLE fTime;
} PEAK, *PPEAK;

typedef struct _TIMECODE_INFO
{
    BOOL bDirection; // true = forward, false = backward
    FLOAT fVelocity;
    BOOL bPaused; // true = paused, false = playing
} TIMECODE_INFO, *PTIMECODE_INFO;

class CTimecode
{
    public:

        CTimecode(DWORD dwBufSize, DOUBLE fSampleRate);
        ~CTimecode(void);

        VOID Analyze(PTIMECODE_INFO pInfo, PFLOAT pBuffer, DWORD dwLength);
        VOID Reset(void);
        VOID Learn(void);

    private:

        BOOL m_bLearn;

        DOUBLE m_Time;

        DWORD m_dwPeaks[2];
        PEAK m_Peaks[2][1024];

        DWORD m_dwZeroPeaks;
        DWORD m_dwHighPeaks;
        DWORD m_dwLowPeaks;

        DOUBLE m_fSampleRate;

        DOUBLE m_fDefaultFreq;
        WORD m_wDefaultPhase; // Default phase represents phase shift during forward playback
};

#endif
